﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  [StorableClass]
  [Item("Sliding Window GP Analyzer", "Base class for concrete sliding window GP analyzers.")]
  public abstract class SlidingWindowAnalyzer<T> : SingleSuccessorOperator, IAnalyzer where T : class,IDataAnalysisProblemData {
    private const string ProblemDataParameterName = "ProblemData";
    private const string ResultCollectionParameterName = "Results";
    private const string FitnessCalculationPartitionParameterName = "FitnessCalculationPartition";
    private const string SlidingWindowSizeParameterName = "Sliding Window Size";
    private const string SlidingWindowStepWidthParameterName = "Sliding Window Step Width";

    private const string SlidingWindowDataResultName = "Sliding Window";

    #region parameter properties
    public IValueLookupParameter<T> ProblemDataParameter {
      get { return (IValueLookupParameter<T>)Parameters[ProblemDataParameterName]; }
    }
    public ILookupParameter<ResultCollection> ResultCollectionParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultCollectionParameterName]; }
    }
    public ILookupParameter<IntRange> FitnessCalculationPartitionParameter {
      get { return (ILookupParameter<IntRange>)Parameters[FitnessCalculationPartitionParameterName]; }
    }

    public IFixedValueParameter<IntValue> SlidingWindowSizeParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[SlidingWindowSizeParameterName]; }
    }
    public IFixedValueParameter<IntValue> SlidingWindowStepWidthParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[SlidingWindowStepWidthParameterName]; }
    }
    #endregion

    #region properties
    public bool EnabledByDefault { get { return false; } }
    public IntValue SlidingWindowSize { get { return SlidingWindowSizeParameter.Value; } }
    public IntValue SlidingWindowStepWidth { get { return SlidingWindowStepWidthParameter.Value; } }
    #endregion


    [StorableConstructor]
    protected SlidingWindowAnalyzer(bool deserializing) : base(deserializing) { }
    protected SlidingWindowAnalyzer(SlidingWindowAnalyzer<T> original, Cloner cloner)
      : base(original, cloner) { }

    protected SlidingWindowAnalyzer() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultCollectionParameterName, "The result collection to store the analysis results."));
      Parameters.Add(new ValueLookupParameter<T>(ProblemDataParameterName, "The problem data on which the symbolic data analysis solution should be evaluated."));
      Parameters.Add(new LookupParameter<IntRange>(FitnessCalculationPartitionParameterName, ""));
      Parameters.Add(new FixedValueParameter<IntValue>(SlidingWindowSizeParameterName, "", new IntValue(1)));
      Parameters.Add(new FixedValueParameter<IntValue>(SlidingWindowStepWidthParameterName, "", new IntValue(1)));
    }

    public override IOperation Apply() {
      if (CheckForUpdate()) {
        var fitnessPartition = FitnessCalculationPartitionParameter.ActualValue;
        if (fitnessPartition.End - fitnessPartition.Start == SlidingWindowSize.Value)
          fitnessPartition.Start += SlidingWindowStepWidth.Value;

        fitnessPartition.End += SlidingWindowStepWidth.Value;
        if (fitnessPartition.End - fitnessPartition.Start > SlidingWindowSize.Value)
          fitnessPartition.End = fitnessPartition.Start + SlidingWindowSize.Value;
      }

      //create and update result
      var results = ResultCollectionParameter.ActualValue;
      SlidingWindowData slidingWindowData;
      if (!results.ContainsKey(SlidingWindowDataResultName)) {
        string targetVariable;
        var classificationProblemData = ProblemDataParameter.ActualValue as IClassificationProblemData;
        var regressionProblemData = ProblemDataParameter.ActualValue as IRegressionProblemData;
        if (classificationProblemData != null) targetVariable = classificationProblemData.TargetVariable;
        else if (regressionProblemData != null) targetVariable = regressionProblemData.TargetVariable;
        else throw new NotSupportedException();

        var targetData = ProblemDataParameter.ActualValue.Dataset.GetDoubleValues(targetVariable);
        slidingWindowData = new SlidingWindowData(FitnessCalculationPartitionParameter.ActualValue, targetData);
        results.Add(new Result(SlidingWindowDataResultName, slidingWindowData));
      } else slidingWindowData = (SlidingWindowData)results[SlidingWindowDataResultName].Value;

      slidingWindowData.SlidingWindowPosition.Start = FitnessCalculationPartitionParameter.ActualValue.Start;
      slidingWindowData.SlidingWindowPosition.End = FitnessCalculationPartitionParameter.ActualValue.End;
      return base.Apply();
    }


    protected abstract bool CheckForUpdate();
  }
}
